/*

 MikMod Sound System

  By Jake Stine of Divine Entertainment (1996-2000)

 Support:
  If you find problems with this code, send mail to:
    air@divent.org

 Distribution / Code rights:
  Use this source code in any fashion you see fit.  Giving me credit where
  credit is due is optional, depending on your own levels of integrity and
  honesty.

 -----------------------------------------
 Module:  LOAD_ULT.C

  Ultratracker (ULT) module loader.  I used to have a couple cool .ULTs, but
  lost them some time ago.  Hope this loader still works! :)

 Portability:
  All systems - all compilers (hopefully)

 If this module is found to not be portable to any particular platform,
 please contact Jake Stine at dracoirs@epix.net (see MIKMOD.TXT for
 more information on contacting the author).

*/

#include <string.h>
#include "mikmod.h"
#include "uniform.h"
#include "..\..\..\..\mikamp\resource.h"


#define ULTS_16BITS     4
#define ULTS_LOOP       8
#define ULTS_REVERSE    16


// Raw ULT header struct:

typedef struct ULTHEADER
{   CHAR  id[16];
    CHAR  songtitle[32];
    UBYTE reserved;
} ULTHEADER;


// Raw ULT sampleinfo struct:

typedef struct ULTSAMPLE
{   CHAR   samplename[32];
    CHAR   dosname[12];
    SLONG  loopstart;
    SLONG  loopend;
    SLONG  sizestart;
    SLONG  sizeend;
    UBYTE  volume;
    UBYTE  flags;
    SWORD  finetune;
} ULTSAMPLE;


typedef struct ULTEVENT
{   UBYTE note,sample,eff,dat1,dat2;
} ULTEVENT;


static const CHAR ULT_Version[] = "Ultra Tracker V1.x";


// =====================================================================================
    BOOL ULT_Test(MMSTREAM *mmfile)
// =====================================================================================
{
    CHAR id[14];

    if (!_mm_read_string(id, 14, mmfile)) return 0;
    return !strncmp(id, "MAS_UTrack_V00", 14);
}

typedef struct ULT_HANDLE
{
    ULTHEADER   mh;
    uint        numtrk;
    UBYTE     **globtrack;
    MM_ALLOC   *allochandle;
} ULT_HANDLE;


// =====================================================================================
static void *ULT_Init(void)
// =====================================================================================
{
    ULT_HANDLE *handle;

    _MM_ALLOCATE(handle, ULT_HANDLE, NULL);
    return handle;
}


// =====================================================================================
static void ULT_Cleanup(ULT_HANDLE *handle)
// =====================================================================================
{
    _mmalloc_close(handle->allochandle);
}


// =====================================================================================
static int ReadUltEvent(ULTEVENT *event, MMSTREAM *mmfile)
// =====================================================================================
{
    UBYTE flag,rep=1;

    flag = _mm_read_UBYTE(mmfile);

    if(flag==0xfc)
    {   rep = _mm_read_UBYTE(mmfile);
        event->note =_mm_read_UBYTE(mmfile);
    } else
        event->note = flag;

    event->sample   =_mm_read_UBYTE(mmfile);
    event->eff      =_mm_read_UBYTE(mmfile);
    event->dat1     =_mm_read_UBYTE(mmfile);
    event->dat2     =_mm_read_UBYTE(mmfile);

    return rep;
}


// =====================================================================================
static BOOL ULT_Load(ULT_HANDLE *handle, UNIMOD *of, MMSTREAM *mmfile)
// =====================================================================================
{
    uint        t,tracks=0;
    uint        u;
    UNISAMPLE  *q;
    ULTSAMPLE   s;
    UBYTE       nos,noc,nop;
    ULTEVENT    ev;

    ULTHEADER  *mh = &handle->mh;

    // try to read module header

    _mm_read_string(mh->id,15,mmfile);
    _mm_read_string(mh->songtitle,32,mmfile);
    mh->reserved = _mm_read_UBYTE(mmfile);

    if(mh->id[14]<'1' || mh->id[14]>'4')
        return 0;

    of->modtype     = _mm_strdup(of->allochandle, ULT_Version);
    of->modtype[17] = mh->id[14] + 2;
    of->initspeed   = 6;
    of->inittempo   = 125;

    // read songtext

    of->comment = ReadComment(of->allochandle, mh->reserved*32, 32, mmfile);

    nos = _mm_read_UBYTE(mmfile);

    if(_mm_feof(mmfile))
    {
        _mmlogd("load_ult> unexpected end of file (in header)");
        return 0;
    }

    of->songname = DupStrTrim(of->allochandle, mh->songtitle, 32);
    of->numsmp   = nos;

    if(!AllocSamples(of,0)) return 0;

    q = of->samples;

    for(t=0; t<nos; t++)
    {   // try to read sample info

        _mm_read_string(s.samplename,32,mmfile);
        _mm_read_string(s.dosname,12,mmfile);
        s.loopstart     =_mm_read_I_ULONG(mmfile);
        s.loopend       =_mm_read_I_ULONG(mmfile);
        s.sizestart     =_mm_read_I_ULONG(mmfile);
        s.sizeend       =_mm_read_I_ULONG(mmfile);
        s.volume        =_mm_read_UBYTE(mmfile);
        s.flags         =_mm_read_UBYTE(mmfile);
        s.finetune      =_mm_read_I_SWORD(mmfile);

        if(_mm_feof(mmfile))
        {
            _mmlogd("load_ult> unexpected end of file (in samples)");
            return 0;
        }

        q->samplename = DupStr(of->allochandle, s.samplename,32);

        q->speed      = 8363;

        if(mh->id[14]>='4')
        {   _mm_read_I_UWORD(mmfile);       // read 1.6 extra info(??) word
            q->speed = s.finetune;
        }

        q->length    = s.sizeend-s.sizestart;
        q->volume    = s.volume/2;
        q->loopstart = s.loopstart;
        q->loopend   = s.loopend;

        q->format    = SF_SIGNED;

        if(s.flags & ULTS_LOOP)
            q->flags |= SL_LOOP;

        if(s.flags & ULTS_16BITS)
        {   q->format     |= SF_16BITS;
            q->loopstart >>= 1;
            q->loopend   >>= 1;
        }

        q++;
    }

    if(!AllocPositions(of,256)) return 0;
    for(t=0; t<256; t++)
        of->positions[t] = _mm_read_UBYTE(mmfile);
    for(t=0; t<256; t++)
        if(of->positions[t]==255) break;

    of->numpos = t;

    noc = _mm_read_UBYTE(mmfile);
    nop = _mm_read_UBYTE(mmfile);

    of->numchn = noc+1;
    of->numpat = nop+1;
    of->numtrk = of->numchn*of->numpat;

		if (of->numchn>=64)
		{
			_mmlogd("load_ult> unexpected channel count (>64)");
			return 0;
		}

    if(!AllocTracks(of)) return 0;
    if(!AllocPatterns(of)) return 0;

    for(u=0; u<of->numchn; u++)
    {   for(t=0; t<of->numpat; t++)
            of->patterns[(t*of->numchn)+u] = tracks++;
    }

    // read pan position table for v1.5 and higher

    if(mh->id[14]>='3')
        for(u=0; u<of->numchn; u++) of->panning[u] = (_mm_read_UBYTE(mmfile) * 16) + PAN_LEFT;

    of->ut = utrk_init(1, handle->allochandle);
    utrk_memory_reset(of->ut);
    utrk_local_memflag(of->ut, PTMEM_PORTAMENTO, UMF_EFFECT);

    if(!(handle->globtrack = (UBYTE **)MikMod_calloc(handle->allochandle, of->numtrk, sizeof(UBYTE *)))) return 0;

    for(t=0; t<of->numtrk; t++)
    {   int rep,s,done;

        utrk_reset(of->ut);
        done = 0;

        while(done < 64)
        {
            rep = ReadUltEvent(&ev,mmfile);
            if(_mm_feof(mmfile))
            {
                _mmlogd("load_ult> unexpected end of file (in patterns)");
                return 0;
            }

            for(s=0; s<rep; s++)
            {   uint  eff;

                utrk_write_inst(of->ut, ev.sample);
                if(ev.note) utrk_write_note(of->ut, ev.note+24);

                eff = ev.eff>>4;

                //  ULT panning effect fixed by Alexander Kerkhove :

                if(eff==0xc) pt_write_effect(of->ut, eff,ev.dat2>>2);
                else if(eff==0xb) pt_write_effect(of->ut, 8,ev.dat2*0xf);
                else pt_write_effect(of->ut, eff,ev.dat2);

                eff = ev.eff&0xf;

                if(eff==0xc) pt_write_effect(of->ut, eff,ev.dat1>>2);
                else if(eff==0xb) pt_write_effect(of->ut, 8,ev.dat1*0xf);
                else pt_write_effect(of->ut, eff,ev.dat1);

                done++;
                utrk_newline(of->ut);
            }
        }

        if(!(of->tracks[t] = utrk_dup_track(of->ut, 0, of->allochandle))) return 0;
        handle->globtrack[t] = utrk_dup_global(of->ut, handle->allochandle);
    }

    // Pattern setup is read, so now process global tracks.
    pt_global_consolidate(of,handle->globtrack);

    {
        long  seekpos = _mm_ftell(mmfile);

        q = of->samples;
        for(t=0; t<of->numsmp; t++, q++)
        {
            q->seekpos = seekpos;
            seekpos   += q->length * ((q->format & SF_16BITS) ? 2 : 1);
        }
    }

    return 1;
}


// =====================================================================================
static CHAR *ULT_LoadTitle(MMSTREAM *mmfile)
// =====================================================================================
{
   CHAR s[32];

   _mm_fseek(mmfile,15,SEEK_SET);
   if(!_mm_read_UBYTES(s,32,mmfile)) return NULL;
   
   return DupStrTrim(NULL, s, 32);
}


MLOADER load_ult =
{   
    "ULT",
    IDS_FAMILY_STRING_ULTRATRACKER,//"Ultratracker",
    0,                              // default FULL STEREO panning
    NULL,
    ULT_Test,
    ULT_Init,
    ULT_Cleanup,
    ULT_Load,
#ifndef _MM_WINAMP_
    ULT_LoadTitle
#endif
};