/*
* ALAC (Apple Lossless Audio Codec) decoder
* Copyright (c) 2006 David Hammerton, Ben Allison
* All rights reserved.
*
* This is the actual decoder.
*
* http://crazney.net/programs/itunes/alac.html
* 
* Permission is hereby granted, free of charge, to any person
* obtaining a copy of this software and associated documentation
* files (the "Software"), to deal in the Software without
* restriction, including without limitation the rights to use,
* copy, modify, merge, publish, distribute, sublicense, and/or
* sell copies of the Software, and to permit persons to whom the
* Software is furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
* OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
* NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
* HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
* FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
* OTHER DEALINGS IN THE SOFTWARE.
*
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "alac_decode.h"
#include "nu/ByteReader.h"
#include "foundation/error.h"

typedef enum
{
	ID_SCE = 0,                                         /* Single Channel Element   */
	ID_CPE = 1,                                         /* Channel Pair Element     */
	ID_CCE = 2,                                         /* Coupling Channel Element */
	ID_LFE = 3,                                         /* LFE Channel Element      */
	ID_DSE = 4,                                         /* not yet supported        */
	ID_PCE = 5,
	ID_FIL = 6,
	ID_END = 7
};

typedef struct alac_decoder_s
{
	const uint8_t *input_buffer;
	ptrdiff_t input_buffer_length_bits;
	size_t input_buffer_bitaccumulator; /* used so we can do arbitary
																			bit reads */
	/* buffers */
	int32_t *predicterror_buffer;
	alac_buffer_t outputsamples_buffer;
	alac_buffer_t uncompressed_bytes_buffer;

	/* stuff from setinfo */
	uint32_t setinfo_max_samples_per_frame; /* 0x1000 = 4096 */    /* max samples per frame? */
	uint8_t setinfo_version; /* 0x00 */
	uint8_t setinfo_sample_size; /* 0x10 */ /* bits per sample */
	uint8_t setinfo_rice_historymult; /* 0x28 */
	uint8_t setinfo_rice_initialhistory; /* 0x0a */
	uint8_t setinfo_rice_kmodifier; /* 0x0e */
	uint8_t setinfo_channels; /* 0x02 */ /* channels? */
	uint16_t setinfo_maxrun; /* 0x00ff, 255 */
	uint32_t setinfo_max_frame_bytes; /* 0x000020e7, 8423 */ /* max sample size?? */
	uint32_t setinfo_bitrate; /* 0x00069fe4, 434148 */ /* bit rate (avarge)?? */
	uint32_t setinfo_sample_rate; /* 0x0000ac44, 44100 */ /* samplerate? */
	/* end setinfo stuff */
} alac_decoder_s;

static int allocate_buffers(alac_decoder_t alac)
{
	unsigned int i;

	alac->predicterror_buffer = (int32_t *)malloc(alac->setinfo_max_samples_per_frame * sizeof(int32_t));
	if (!alac->predicterror_buffer)
		return NErr_OutOfMemory;


	/* --- */

	alac->outputsamples_buffer = (alac_buffer_t)calloc(sizeof(int32_t *), alac->setinfo_channels);
	if (!alac->outputsamples_buffer)
		return NErr_OutOfMemory;
	for (i=0;i<alac->setinfo_channels;i++)
	{
		alac->outputsamples_buffer[i] = (int32_t *)malloc(alac->setinfo_max_samples_per_frame * sizeof(int32_t));
		if (!alac->outputsamples_buffer[i])
			return NErr_OutOfMemory;
	}

	/* --- */

	alac->uncompressed_bytes_buffer = (alac_buffer_t)calloc(sizeof(int32_t *), alac->setinfo_channels);
	if (!alac->uncompressed_bytes_buffer)
		return NErr_OutOfMemory;

	for (i=0;i<alac->setinfo_channels;i++)
	{
		alac->uncompressed_bytes_buffer[i] = (int32_t *)malloc(alac->setinfo_max_samples_per_frame * sizeof(int32_t));
		if (!alac->uncompressed_bytes_buffer[i])
			return NErr_OutOfMemory;
	}

	return NErr_Success;
}

static void deallocate_buffers(alac_decoder_t alac)
{
	unsigned int i;

	free(alac->predicterror_buffer);

	if (alac->outputsamples_buffer)
	{
		for (i=0;i<alac->setinfo_channels;i++)
			free(alac->outputsamples_buffer[i]);
	}
	free(alac->outputsamples_buffer);

	if (alac->uncompressed_bytes_buffer)
	{
		for (i=0;i<alac->setinfo_channels;i++)
			free(alac->uncompressed_bytes_buffer[i]);
	}
	free(alac->uncompressed_bytes_buffer);

}

int alac_configure(alac_decoder_t alac, const void *configuration_data, size_t configuration_length)
{
	bytereader_value_t byte_reader;
	if (configuration_length < 24)
		return NErr_NeedMoreData;

	bytereader_init(&byte_reader, configuration_data, configuration_length);

	alac->setinfo_max_samples_per_frame = bytereader_read_u32_be(&byte_reader); /* buffer size / 2 ? */
	alac->setinfo_version = bytereader_read_u8(&byte_reader);
	alac->setinfo_sample_size = bytereader_read_u8(&byte_reader);

	alac->setinfo_rice_historymult = bytereader_read_u8(&byte_reader);
	alac->setinfo_rice_initialhistory = bytereader_read_u8(&byte_reader);
	alac->setinfo_rice_kmodifier = bytereader_read_u8(&byte_reader);
	alac->setinfo_channels = bytereader_read_u8(&byte_reader);

	alac->setinfo_maxrun = bytereader_read_u16_be(&byte_reader);
	alac->setinfo_max_frame_bytes = bytereader_read_u32_be(&byte_reader);
	alac->setinfo_bitrate = bytereader_read_u32_be(&byte_reader);
	alac->setinfo_sample_rate = bytereader_read_u32_be(&byte_reader);

	return allocate_buffers(alac);
}

/* stream reading */

/* supports reading 1 to 24 bits, in big endian format */
static uint32_t readbits_24(alac_decoder_t alac, int bits)
{
	uint32_t result;
	int new_accumulator;

	result = (alac->input_buffer[0] << 16) |
		(alac->input_buffer[1] << 8) |
		(alac->input_buffer[2]);

	/* shift left by the number of bits we've already read,
	* so that the top 'n' bits of the 24 bits we read will
	* be the return bits */
	result = result << alac->input_buffer_bitaccumulator;

	result = result & 0x00ffffff;

	/* and then only want the top 'n' bits from that, where
	* n is 'bits' */
	result = result >> (24 - bits);

	new_accumulator = (alac->input_buffer_bitaccumulator + bits);

	/* increase the buffer pointer if we've read over n bytes. */
	alac->input_buffer += (new_accumulator >> 3);
	alac->input_buffer_length_bits -= bits;

	/* and the remainder goes back into the bit accumulator */
	alac->input_buffer_bitaccumulator = (new_accumulator & 7);

	return result;
}

/* supports reading 1 to 16 bits, in big endian format */
static uint32_t readbits_16(alac_decoder_t alac, int bits)
{
	uint32_t result;
	int new_accumulator;

	result = (alac->input_buffer[0] << 8) |
		(alac->input_buffer[1]);

	/* shift left by the number of bits we've already read,
	* so that the top 'n' bits of the 16 bits we read will
	* be the return bits */
	result = result << alac->input_buffer_bitaccumulator;

	result = result & 0x0000ffff;

	/* and then only want the top 'n' bits from that, where
	* n is 'bits' */
	result = result >> (16 - bits);

	new_accumulator = (alac->input_buffer_bitaccumulator + bits);

	/* increase the buffer pointer if we've read over n bytes. */
	alac->input_buffer += (new_accumulator >> 3);
	alac->input_buffer_length_bits -= bits;

	/* and the remainder goes back into the bit accumulator */
	alac->input_buffer_bitaccumulator = (new_accumulator & 7);

	return result;
}


/* supports reading 1 to 8 bits, in big endian format */
static uint32_t readbits_8(alac_decoder_t alac, int bits)
{
	uint32_t result;
	int new_accumulator;

	result = (alac->input_buffer[0]);

	/* shift left by the number of bits we've already read,
	* so that the top 'n' bits of the 8 bits we read will
	* be the return bits */
	result = result << alac->input_buffer_bitaccumulator;

	result = result & 0x000000ff;

	/* and then only want the top 'n' bits from that, where
	* n is 'bits' */
	result = result >> (8 - bits);

	new_accumulator = (alac->input_buffer_bitaccumulator + bits);

	/* increase the buffer pointer if we've read over n bytes. */
	alac->input_buffer += (new_accumulator >> 3);
	alac->input_buffer_length_bits -= bits;

	/* and the remainder goes back into the bit accumulator */
	alac->input_buffer_bitaccumulator = (new_accumulator & 7);

	return result;
}

/* supports reading 1 to 32 bits, in big endian format */
static uint32_t readbits(alac_decoder_t alac, int bits)
{
	int32_t result = 0;

	int reqbits = bits + alac->input_buffer_bitaccumulator;

	if (bits > 16)
	{
		bits -= 16;
		if (reqbits > 16)
			result = readbits_24(alac, 16) << bits;
		else
			result = readbits_8(alac, 16) << bits;
	}

	reqbits = bits + alac->input_buffer_bitaccumulator;
	if (reqbits>16)
		result |= readbits_24(alac, bits);
	else if (reqbits>8)
		result |= readbits_16(alac, bits);
	else
		result |= readbits_8(alac, bits);

	return result;
}

/* reads a single bit */
static int readbit(alac_decoder_t alac)
{
	int result;
	int new_accumulator;

	result = alac->input_buffer[0];

	result = result << alac->input_buffer_bitaccumulator;

	result = result >> 7 & 1;

	new_accumulator = (alac->input_buffer_bitaccumulator + 1);

	alac->input_buffer += (new_accumulator / 8);

	alac->input_buffer_bitaccumulator = (new_accumulator % 8);
	alac->input_buffer_length_bits--;

	return result;
}

static void unreadbits(alac_decoder_t alac, int bits)
{
	int new_accumulator = (alac->input_buffer_bitaccumulator - bits);

	alac->input_buffer += (new_accumulator >> 3);
	alac->input_buffer_length_bits += bits;

	alac->input_buffer_bitaccumulator = (new_accumulator & 7);
	if (alac->input_buffer_bitaccumulator < 0)
		alac->input_buffer_bitaccumulator *= -1;
}

/* various implementations of count_leading_zero:
* the first one is the original one, the simplest and most
* obvious for what it's doing. never use this.
* then there are the asm ones. fill in as necessary
* and finally an unrolled and optimised c version
* to fall back to
*/
#if 0 
/* hideously inefficient. could use a bitmask search,
* alternatively bsr on x86,
*/
static int count_leading_zeros(int32_t input)
{
	int i = 0;
	while (!(0x80000000 & input) && i < 32)
	{
		i++;
		input = input << 1;
	}
	return i;
}
#elif defined(__GNUC__) && defined(__arm__) && !defined(__thumb__)
/* for some reason the unrolled version (below) is
* actually faster than this. yay intel!
*/
static int count_leading_zeros(int input)
{
	int output = 0;
	__asm("clz %[result], %[value]\n"
	: [result] "=r" (output)
		: [value] "r" (input));
	return output;
}
#elif defined(__GNUC__) && (defined(_X86) || defined(__i386) || defined(i386)) 
/* for some reason the unrolled version (below) is
* actually faster than this. yay intel!
*/
static int count_leading_zeros(int input)
{
	int output = 0;
	if (!input) return 32;
	__asm("bsr %1, %0\n"
	: "=r" (output)
		: "r" (input));
	return (0x1f - output);
}
#elif defined(_MSC_VER) && defined(_M_IX86)
static int count_leading_zeros(int input)
{
	int output = 0;
	if (!input) return 32;
	__asm
	{
		mov eax, input;
		mov edx, 0x1f;
		bsr ecx, eax;
		sub edx, ecx;
		mov output, edx;
	}
	return output;
}
#else
//#warning using generic count leading zeroes. You may wish to write one for your CPU / compiler
static int count_leading_zeros(int input)
{
	int output = 0;
	int curbyte = 0;

	curbyte = input >> 24;
	if (curbyte) goto found;
	output += 8;

	curbyte = input >> 16;
	if (curbyte & 0xff) goto found;
	output += 8;

	curbyte = input >> 8;
	if (curbyte & 0xff) goto found;
	output += 8;

	curbyte = input;
	if (curbyte & 0xff) goto found;
	output += 8;

	return output;

found:
	if (!(curbyte & 0xf0))
	{
		output += 4;
	}
	else
		curbyte >>= 4;

	if (curbyte & 0x8)
		return output;
	if (curbyte & 0x4)
		return output + 1;
	if (curbyte & 0x2)
		return output + 2;
	if (curbyte & 0x1)
		return output + 3;

	/* shouldn't get here: */
	return output + 4;
}
#endif

#define RICE_THRESHOLD 8 // maximum number of bits for a rice prefix.

static int32_t entropy_decode_value(alac_decoder_t alac, int readSampleSize, int k, int rice_kmodifier_mask)
{
	int32_t x = 0; // decoded value

	// read x, number of 1s before 0 represent the rice value.
	while (x <= RICE_THRESHOLD && readbit(alac))
	{
		x++;
	}

	if (x > RICE_THRESHOLD)
	{
		// read the number from the bit stream (raw value)
		int32_t value;

		value = readbits(alac, readSampleSize);

		// mask value
		value &= (((uint32_t)0xffffffff) >> (32 - readSampleSize));

		x = value;
	}
	else
	{
		if (k != 1)
		{
			int extraBits = readbits(alac, k);

			// x = x * (2^k - 1)
			x *= (((1 << k) - 1) & rice_kmodifier_mask);

			if (extraBits > 1)
				x += extraBits - 1;
			else
				unreadbits(alac, 1);
		}
	}

	return x;
}

static void entropy_rice_decode(alac_decoder_t alac, int32_t *outputBuffer, int outputSize, int readSampleSize, int rice_initialhistory, int rice_kmodifier, int rice_historymult, int rice_kmodifier_mask)
{
	int				outputCount;
	int				history = rice_initialhistory;
	int				signModifier = 0;

	for (outputCount = 0; outputCount < outputSize; outputCount++)
	{
		int32_t		decodedValue;
		int32_t		finalValue;
		int32_t		k;

		k = 31 - rice_kmodifier - count_leading_zeros((history >> 9) + 3);

		if (k < 0) k += rice_kmodifier;
		else k = rice_kmodifier;

		// note: don't use rice_kmodifier_mask here (set mask to 0xFFFFFFFF)
		decodedValue = entropy_decode_value(alac, readSampleSize, k, 0xFFFFFFFF);

		decodedValue += signModifier;
		finalValue = (decodedValue + 1) / 2; // inc by 1 and shift out sign bit
		if (decodedValue & 1) // the sign is stored in the low bit
			finalValue *= -1;

		outputBuffer[outputCount] = finalValue;

		signModifier = 0;

		// update history
		history += (decodedValue * rice_historymult)
			- ((history * rice_historymult) >> 9);

		if (decodedValue > 0xFFFF)
			history = 0xFFFF;

		// special case, for compressed blocks of 0
		if ((history < 128) && (outputCount + 1 < outputSize))
		{
			int32_t		blockSize;

			signModifier = 1;

			k = count_leading_zeros(history) + ((history + 16) / 64) - 24;

			// note: blockSize is always 16bit
			blockSize = entropy_decode_value(alac, 16, k, rice_kmodifier_mask);

			// got blockSize 0s
			if (blockSize > 0)
			{
				memset(&outputBuffer[outputCount + 1], 0, blockSize * sizeof(*outputBuffer));
				outputCount += blockSize;
			}

			if (blockSize > 0xFFFF)
				signModifier = 0;

			history = 0;
		}
	}
}


#define SIGN_EXTENDED32(val, bits) ((val << (32 - bits)) >> (32 - bits))

#define SIGN_ONLY(v) \
	((v < 0) ? (-1) : \
	((v > 0) ? (1) : \
	(0)))

static void predictor_decompress_fir_adapt(const int32_t *error_buffer,
	int32_t *buffer_out,
	int output_size,
	int readsamplesize,
	int16_t *predictor_coef_table,
	int predictor_coef_num,
	int predictor_quantitization)
{
	int i;

	/* first sample always copies */
	*buffer_out = *error_buffer;

	if (!predictor_coef_num)
	{
		if (output_size <= 1) return ;
		memcpy(buffer_out + 1, error_buffer + 1, (output_size - 1) * 4);
		return ;
	}

	if (predictor_coef_num == 0x1f) /* 11111 - max value of predictor_coef_num */
	{ /* second-best case scenario for fir decompression,
		* error describes a small difference from the previous sample only
		*/
		if (output_size <= 1) return ;
		for (i = 0; i < output_size - 1; i++)
		{
			int32_t prev_value;
			int32_t error_value;

			prev_value = buffer_out[i];
			error_value = error_buffer[i + 1];
			buffer_out[i + 1] = SIGN_EXTENDED32((prev_value + error_value), readsamplesize);
		}
		return ;
	}

	/* read warm-up samples */
	if (predictor_coef_num > 0)
	{
		int i;
		for (i = 0; i < predictor_coef_num; i++)
		{
			int32_t val;

			val = buffer_out[i] + error_buffer[i + 1];

			val = SIGN_EXTENDED32(val, readsamplesize);

			buffer_out[i + 1] = val;
		}
	}

#if 0
	/* 4 and 8 are very common cases (the only ones i've seen). these
	* should be unrolled and optimised
	*/
	if (predictor_coef_num == 4)
	{
		/* FIXME: optimised general case */
		return ;
	}

	if (predictor_coef_table == 8)
	{
		/* FIXME: optimised general case */
		return ;
	}
#endif


	/* general case */
	if (predictor_coef_num > 0)
	{
		for (i = predictor_coef_num + 1;
			i < output_size;
			i++)
		{
			int j;
			int sum = 0;
			int outval;
			int error_val = error_buffer[i];

			for (j = 0; j < predictor_coef_num; j++)
			{
				sum += (buffer_out[predictor_coef_num - j] - buffer_out[0]) *
					predictor_coef_table[j];
			}

			outval = (1 << (predictor_quantitization - 1)) + sum;
			outval = outval >> predictor_quantitization;
			outval = outval + buffer_out[0] + error_val;
			outval = SIGN_EXTENDED32(outval, readsamplesize);

			buffer_out[predictor_coef_num + 1] = outval;

			if (error_val > 0)
			{
				int predictor_num = predictor_coef_num - 1;

				while (predictor_num >= 0 && error_val > 0)
				{
					int val = buffer_out[0] - buffer_out[predictor_coef_num - predictor_num];
					int sign = SIGN_ONLY(val);

					predictor_coef_table[predictor_num] -= sign;

					val *= sign; /* absolute value */

					error_val -= ((val >> predictor_quantitization) *
						(predictor_coef_num - predictor_num));

					predictor_num--;
				}
			}
			else if (error_val < 0)
			{
				int predictor_num = predictor_coef_num - 1;

				while (predictor_num >= 0 && error_val < 0)
				{
					int val = buffer_out[0] - buffer_out[predictor_coef_num - predictor_num];
					int sign = - SIGN_ONLY(val);

					predictor_coef_table[predictor_num] -= sign;

					val *= sign; /* neg value */

					error_val -= ((val >> predictor_quantitization) *
						(predictor_coef_num - predictor_num));

					predictor_num--;
				}
			}

			buffer_out++;
		}
	}
}

static void weight_stereo(alac_buffer_t buffer, const alac_buffer_t uncompressed_bytes_buffer, int uncompressed_bytes, int numchannels, int numsamples, uint8_t interlacing_shift, uint8_t interlacing_leftweight)
{
	int i;
	if (numsamples <= 0) return ;

	/* weighted interlacing */
	if (interlacing_leftweight)
	{
		if (uncompressed_bytes)
		{
			for (i = 0; i < numsamples; i++)
			{
				int32_t difference, midright;
				int32_t left;
				int32_t right;

				midright = buffer[0][i];
				difference = buffer[1][i];

				right = midright - ((difference * interlacing_leftweight) >> interlacing_shift);
				left = right + difference;

				if (uncompressed_bytes)
				{
					uint32_t mask = ~(0xFFFFFFFF << (uncompressed_bytes * 8));
					left <<= (uncompressed_bytes * 8);
					right <<= (uncompressed_bytes * 8);

					left |= uncompressed_bytes_buffer[0][i] & mask;
					right |= uncompressed_bytes_buffer[1][i] & mask;
				}

				buffer[0][i] = left;
				buffer[1][i] = right;			
			}
		}
		else
		{
			for (i = 0; i < numsamples; i++)
			{
				int32_t difference, midright;
				int32_t left;
				int32_t right;

				midright = buffer[0][i];
				difference = buffer[1][i];

				right = midright - ((difference * interlacing_leftweight) >> interlacing_shift);
				left = right + difference;

				buffer[0][i] = left;
				buffer[1][i] = right;
			}
		}
		return ;
	}
	else if (uncompressed_bytes)
	{
		/* otherwise basic interlacing took place */
		for (i = 0; i < numsamples; i++)
		{
			int32_t left, right;
			uint32_t mask;
			left = buffer[0][i];
			right = buffer[1][i];

			mask = ~(0xFFFFFFFF << (uncompressed_bytes * 8));
			left <<= (uncompressed_bytes * 8);
			right <<= (uncompressed_bytes * 8);

			left |= uncompressed_bytes_buffer[0][i] & mask;
			right |= uncompressed_bytes_buffer[1][i] & mask;


			buffer[0][i] = left;
			buffer[1][i] = right;		
		}
	}
}

static int decode_mono(alac_decoder_t alac,	alac_buffer_t * const outbuffer, size_t *frames)
{
	int hassize;
	int isnotcompressed;
	int readsamplesize;

	int uncompressed_bytes;
	int ricemodifier;
	int32_t outputsamples = alac->setinfo_max_samples_per_frame;

	/* 2^result = something to do with output waiting.
	* perhaps matters if we read > 1 frame in a pass?
	*/
	uint32_t element_instance_tag = readbits(alac, 4);
	readbits(alac, 12); /* unused 12 bits */
	hassize = readbits(alac, 1); /* the output sample size is stored soon */
	uncompressed_bytes = readbits(alac, 2); /* number of bytes in the (compressed) stream that are not compressed */
	isnotcompressed = readbits(alac, 1); /* whether the frame is compressed */

	if (hassize)
	{
		/* now read the number of samples,
		* as a 32bit integer */
		outputsamples = readbits(alac, 32);
	}

	*frames = outputsamples;

	readsamplesize = alac->setinfo_sample_size - (uncompressed_bytes * 8);

	if (!isnotcompressed)
	{ /* so it is compressed */
		int16_t predictor_coef_table[32];
		int predictor_coef_num;
		int prediction_type;
		int prediction_quantitization;
		int i;

		/* skip 16 bits, not sure what they are. seem to be used in
		* two channel case */
		unsigned int mix_bits = readbits(alac, 8);
		unsigned int mix_res = readbits(alac, 8);

		prediction_type = readbits(alac, 4);
		prediction_quantitization = readbits(alac, 4);

		ricemodifier = readbits(alac, 3);
		predictor_coef_num = readbits(alac, 5);

		/* read the predictor table */
		for (i = 0; i < predictor_coef_num; i++)
		{
			predictor_coef_table[i] = (int16_t)readbits(alac, 16);
		}

		if (uncompressed_bytes)
		{
			int i;
			for (i = 0; i < outputsamples; i++)
			{
				alac->uncompressed_bytes_buffer[0][i] = readbits(alac, uncompressed_bytes * 8);
			}
		}

		entropy_rice_decode(alac,
			alac->predicterror_buffer,
			outputsamples,
			readsamplesize,
			alac->setinfo_rice_initialhistory,
			alac->setinfo_rice_kmodifier,
			ricemodifier * alac->setinfo_rice_historymult / 4,
			(1 << alac->setinfo_rice_kmodifier) - 1);

		if (prediction_type == 0)
		{ /* adaptive fir */
			predictor_decompress_fir_adapt(alac->predicterror_buffer,
				alac->outputsamples_buffer[0],
				outputsamples,
				readsamplesize,
				predictor_coef_table,
				predictor_coef_num,
				prediction_quantitization);
		}
		else
		{
			fprintf(stderr, "FIXME: unhandled predicition type: %i\n", prediction_type);
			/* i think the only other prediction type (or perhaps this is just a
			* boolean?) runs adaptive fir twice.. like:
			* predictor_decompress_fir_adapt(predictor_error, tempout, ...)
			* predictor_decompress_fir_adapt(predictor_error, outputsamples ...)
			* little strange..
			*/
		}

	}
	else
	{ /* not compressed, easy case */
		int i;
		for (i = 0; i < outputsamples; i++)
		{
			int32_t audiobits = readbits(alac, alac->setinfo_sample_size);

			audiobits = SIGN_EXTENDED32(audiobits, alac->setinfo_sample_size);

			alac->outputsamples_buffer[0][i] = audiobits;
		}

		uncompressed_bytes = 0; // always 0 for uncompressed
	}

	if (uncompressed_bytes)
	{
		int i;
		for (i = 0; i < outputsamples; i++)
		{
			int32_t sample = alac->outputsamples_buffer[0][i];

			uint32_t mask;
			sample = sample << (uncompressed_bytes * 8);
			mask = ~(0xFFFFFFFF << (uncompressed_bytes * 8));
			sample |= alac->uncompressed_bytes_buffer[0][i] & mask;


			alac->outputsamples_buffer[0][i] = sample;
		}
	}
	*outbuffer = alac->outputsamples_buffer;
	return NErr_Success;

}

int decode_stereo(alac_decoder_t alac, alac_buffer_t * const outbuffer, size_t *frames)
{
	int hassize;
	int isnotcompressed;
	int readsamplesize;

	int uncompressed_bytes;

	uint8_t interlacing_shift;
	uint8_t interlacing_leftweight;
	int32_t outputsamples = alac->setinfo_max_samples_per_frame;

	/* 2^result = something to do with output waiting.
	* perhaps matters if we read > 1 frame in a pass?
	*/
	readbits(alac, 4);
	readbits(alac, 12); /* unknown, skip 12 bits */
	hassize = readbits(alac, 1); /* the output sample size is stored soon */
	uncompressed_bytes = readbits(alac, 2); /* the number of bytes in the (compressed) stream that are not compressed */
	isnotcompressed = readbits(alac, 1); /* whether the frame is compressed */

	if (hassize)
	{
		/* now read the number of samples,
		* as a 32bit integer */
		outputsamples = readbits(alac, 32);
	}

	*frames = outputsamples;

	readsamplesize = alac->setinfo_sample_size - (uncompressed_bytes * 8) + 1;

	if (!isnotcompressed)
	{ /* compressed */
		int16_t predictor_coef_table_a[32];
		int predictor_coef_num_a;
		int prediction_type_a;
		int prediction_quantitization_a;
		int ricemodifier_a;

		int16_t predictor_coef_table_b[32];
		int predictor_coef_num_b;
		int prediction_type_b;
		int prediction_quantitization_b;
		int ricemodifier_b;

		int i;

		interlacing_shift = readbits(alac, 8);
		interlacing_leftweight = readbits(alac, 8);

		/******** channel 1 ***********/
		prediction_type_a = readbits(alac, 4);
		prediction_quantitization_a = readbits(alac, 4);

		ricemodifier_a = readbits(alac, 3);
		predictor_coef_num_a = readbits(alac, 5);

		/* read the predictor table */
		for (i = 0; i < predictor_coef_num_a; i++)
		{
			predictor_coef_table_a[i] = (int16_t)readbits(alac, 16);
		}

		/******** channel 2 *********/
		prediction_type_b = readbits(alac, 4);
		prediction_quantitization_b = readbits(alac, 4);

		ricemodifier_b = readbits(alac, 3);
		predictor_coef_num_b = readbits(alac, 5);

		/* read the predictor table */
		for (i = 0; i < predictor_coef_num_b; i++)
		{
			predictor_coef_table_b[i] = (int16_t)readbits(alac, 16);
		}

		/*********************/
		if (uncompressed_bytes)
		{ /* see mono case */
			int i;
			for (i = 0; i < outputsamples; i++)
			{
				alac->uncompressed_bytes_buffer[0][i] = readbits(alac, uncompressed_bytes * 8);
				alac->uncompressed_bytes_buffer[1][i] = readbits(alac, uncompressed_bytes * 8);
			}
		}

		/* channel 1 */
		entropy_rice_decode(alac,
			alac->predicterror_buffer,
			outputsamples,
			readsamplesize,
			alac->setinfo_rice_initialhistory,
			alac->setinfo_rice_kmodifier,
			ricemodifier_a * alac->setinfo_rice_historymult / 4,
			(1 << alac->setinfo_rice_kmodifier) - 1);

		if (prediction_type_a == 0)
		{ /* adaptive fir */
			predictor_decompress_fir_adapt(alac->predicterror_buffer,
				alac->outputsamples_buffer[0],
				outputsamples,
				readsamplesize,
				predictor_coef_table_a,
				predictor_coef_num_a,
				prediction_quantitization_a);
		}
		else
		{ /* see mono case */
			fprintf(stderr, "FIXME: unhandled predicition type: %i\n", prediction_type_a);
		}

		/* channel 2 */
		entropy_rice_decode(alac,
			alac->predicterror_buffer,
			outputsamples,
			readsamplesize,
			alac->setinfo_rice_initialhistory,
			alac->setinfo_rice_kmodifier,
			ricemodifier_b * alac->setinfo_rice_historymult / 4,
			(1 << alac->setinfo_rice_kmodifier) - 1);

		if (prediction_type_b == 0)
		{ /* adaptive fir */
			predictor_decompress_fir_adapt(alac->predicterror_buffer,
				alac->outputsamples_buffer[1],
				outputsamples,
				readsamplesize,
				predictor_coef_table_b,
				predictor_coef_num_b,
				prediction_quantitization_b);
		}
		else
		{
			fprintf(stderr, "FIXME: unhandled predicition type: %i\n", prediction_type_b);
		}
	}
	else
	{ /* not compressed, easy case */

		int i;
		for (i = 0; i < outputsamples; i++)
		{
			int32_t audiobits_a, audiobits_b;

			audiobits_a = readbits(alac, alac->setinfo_sample_size);
			audiobits_b = readbits(alac, alac->setinfo_sample_size);

			audiobits_a = SIGN_EXTENDED32(audiobits_a, alac->setinfo_sample_size);
			audiobits_b = SIGN_EXTENDED32(audiobits_b, alac->setinfo_sample_size);

			alac->outputsamples_buffer[0][i] = audiobits_a;
			alac->outputsamples_buffer[1][i] = audiobits_b;
		}

		uncompressed_bytes = 0; // always 0 for uncompressed
		interlacing_shift = 0;
		interlacing_leftweight = 0;
	}

	weight_stereo(alac->outputsamples_buffer, alac->uncompressed_bytes_buffer, uncompressed_bytes, alac->setinfo_channels, outputsamples, interlacing_shift, interlacing_leftweight);
	*outbuffer = alac->outputsamples_buffer;
	return NErr_Success;

}

int alac_decode(alac_decoder_t alac, const void *inbuffer, size_t inbuffer_length, alac_buffer_t * const outbuffer, size_t *frames)
{
	int ret;
	int element;
	ptrdiff_t number_of_bits;
	size_t chan=0;
	int32_t outputsamples = alac->setinfo_max_samples_per_frame;

	/* setup the stream */
	number_of_bits = inbuffer_length*8;
	if (number_of_bits < 0 || (size_t)number_of_bits < inbuffer_length*8)
		return NErr_IntegerOverflow;
	alac->input_buffer = (const uint8_t *)inbuffer;
	alac->input_buffer_length_bits = number_of_bits;
	alac->input_buffer_bitaccumulator = 0;

	while (alac->input_buffer_length_bits >= 3)
	{
		element = readbits(alac, 3);

		switch (element)
		{
		case ID_SCE:  /* 1 channel */
			ret = decode_mono(alac, &outbuffer[chan], frames);
			if (ret != NErr_Success)
				return ret;
			chan++;
			break;
	
		case ID_CPE:  /* 2 channels */
			ret = decode_stereo(alac, &outbuffer[chan], frames);
			if (ret != NErr_Success)
				return ret;
			chan+=2;
			break;
	
		case ID_END:
			return NErr_Success;
		default:
			return NErr_NotImplemented;

			
		}

		// align stream
		//number_of_bits = alac->input_buffer_length_bits & 7;
		//if (number_of_bits)
			//readbits(alac, number_of_bits);
	}
	return NErr_Success;
}

int alac_create(alac_decoder_t *decoder)
{
	alac_decoder_t newfile = (alac_decoder_t)calloc(sizeof(alac_decoder_s), 1);
	if (!newfile)
		return NErr_OutOfMemory;
	*decoder = (alac_decoder_t)newfile;

	return NErr_Success;
}

void alac_destroy(alac_decoder_t alac)
{
	deallocate_buffers(alac);
	free(alac);
}

int alac_get_information(alac_decoder_t alac, uint32_t *sample_rate, uint8_t *channels, uint8_t *bps)
{
	*sample_rate = alac->setinfo_sample_rate;
	*channels = alac->setinfo_channels;
	*bps = alac->setinfo_sample_size;
	return NErr_Success;
}